==> furiosa_llm-2024.2.0-py3-none-any.whl: furiosa_llm/models/config_types.py <==
import abc
from enum import Enum
import json
import os
from typing import Literal, Optional, Sequence, Tuple, Union

from pydantic import BaseModel, RootModel
from pydantic.dataclasses import dataclass


class LLMBackend(Enum):
    """The backend implementation to run forward() of a model for the LLM."""

    # FIXME: In order to increase the code consistency, use the capital letter for the enum value
    TORCH_PIPELINE_RUNNER = "torch_pipeline_runner"
    FURIOSA_RT_CUDA = "furiosa_rt_cuda"
    FURIOSA_RT_NPU = "furiosa_rt_npu"
    FURIOSA_RT_V2 = "furiosa_rt_v2"
    MOCK_BACKEND_V2 = "mock_backend_v2"

    def is_parallelism_supported(self) -> bool:
        """Tensor Parallelism/Pipeline Parallelism supports"""
        return self in (
            LLMBackend.TORCH_PIPELINE_RUNNER,
            LLMBackend.FURIOSA_RT_CUDA,
            LLMBackend.FURIOSA_RT_NPU,
            LLMBackend.FURIOSA_RT_V2,
            LLMBackend.MOCK_BACKEND_V2,
        )

    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str):
            return cls[value.upper()]


@dataclass
class SchedulerConfig:
    """
    * version: Version of the `SchedulerConfig`.
    * npu_queue_limit: Maximum number of tasks that can be queued in the hardward
    * max_processing_samples: Maximum number of samples that can be processed by the scheduler
    * spare_blocks_ratio: Ratio of spare blocks that are reserved by scheduler. Smaller value will force the scheduler to use dram aggressively
    * is_offline: If True, use strategies optimzed for offline scenario
    * prefill_chunk_size: Chunk size used for chunked prefill. If the value is `None`, chunked prefill is not used.
    """

    npu_queue_limit: int = 2
    max_processing_samples: int = 65536
    spare_blocks_ratio: float = 0.2
    is_offline: bool = False
    prefill_chunk_size: Optional[int] = None

    # custom comparator to handle float comparison
    def __eq__(self, other):
        return (
            self.npu_queue_limit == other.npu_queue_limit
            and self.max_processing_samples == other.max_processing_samples
            and abs(self.spare_blocks_ratio - other.spare_blocks_ratio) < 1e-6
            and self.is_offline == other.is_offline
        )

    def export(self, path: Union[str, os.PathLike]):
        with open(path, "w") as f:
            f.write(RootModel[SchedulerConfig](self).model_dump_json(indent=4))

    @classmethod
    def load(cls, path: Union[str, os.PathLike]) -> "SchedulerConfig":
        with open(path) as f:
            o = json.load(f)
            return SchedulerConfig(**o)


# FIXME: remove this after fixing generator to support new bucket.
class DeprecatedBucket(BaseModel):
    batch_size: int
    attention_size: int

    def __init__(self, batch_size: int, attention_size: int):
        super(DeprecatedBucket, self).__init__(batch_size=batch_size, attention_size=attention_size)


class Bucket(BaseModel):
    batch_size: int  # batch size, batch size must be a multiple of 2 now.
    attention_size: int
    kv_cache_size: int

    @property
    def input_ids_size(self) -> int:
        return self.attention_size - self.kv_cache_size

    def __init__(self, batch_size: int, attention_size: int, kv_cache_size: int):
        super(Bucket, self).__init__(
            batch_size=batch_size, attention_size=attention_size, kv_cache_size=kv_cache_size
        )

    @classmethod
    def prefill(cls, batch_size: int, attention_size: int) -> "Bucket":
        return cls(batch_size, attention_size, 0)

    @classmethod
    def decode(cls, batch_size: int, attention_size: int) -> "Bucket":
        return cls(batch_size, attention_size, attention_size - 1)

    @property
    def is_prefill(self) -> bool:
        return self.kv_cache_size == 0

    @property
    def is_decode(self) -> bool:
        return self.kv_cache_size > 0

    def __hash__(self) -> int:
        return hash((str(self.__class__), self.batch_size, self.attention_size))


class BucketConfig(abc.ABC): ...


@dataclass
class ManualBucketConfig(BucketConfig):
    prefill_buckets: Sequence[Tuple[int, int]]
    decode_buckets: Optional[Sequence[Tuple[int, int]]] = None


@dataclass
class MinimalBucketConfig(BucketConfig):
    max_seq_len: int


class PagedAttentionConfig(BaseModel):
    """Paged attention configuration.

    Attributes:
        num_blocks (int): The maximum number of blocks that each k/v storage per layer can store
        block_size (int): The maximum number of tokens that can be stored in a single paged attention block
        padding_block_idx (int|None): Padding block's index. This will be used for optimization.
    """

    num_blocks: int
    block_size: int

    # Padding block's index. This will be used for optimization.
    padding_block_idx: Optional[int] = None

    def __init__(self, num_blocks: int, block_size: int, padding_block_idx: Optional[int] = None):
        assert num_blocks > 0
        assert block_size > 0
        super(PagedAttentionConfig, self).__init__(
            num_blocks=num_blocks, block_size=block_size, padding_block_idx=padding_block_idx
        )


class KvCacheSharingAcrossBeamsConfig(BaseModel):
    beam_width: int
    max_new_tokens: int

    def __init__(self, beam_width: int, max_new_tokens: int):
        assert beam_width > 0
        assert max_new_tokens > 0
        super(KvCacheSharingAcrossBeamsConfig, self).__init__(
            beam_width=beam_width, max_new_tokens=max_new_tokens
        )


@dataclass
class GeneratorConfig:
    position_id_pad: int
    buckets: Sequence[Bucket]
    model_qname: str  # qualified name of the model (module + class)
    paged_attention_config: Optional[PagedAttentionConfig]
    packing_type: Literal["IDENTITY"]
    kv_cache_sharing_across_beams_config: Optional[KvCacheSharingAcrossBeamsConfig]
    num_speculative_tokens: Optional[int]

    def export(self, path: Union[str, os.PathLike]):
        with open(path, "w") as f:
            f.write(RootModel[GeneratorConfig](self).model_dump_json(indent=4))

    @classmethod
    def load(cls, path: Union[str, os.PathLike]) -> "GeneratorConfig":
        with open(path) as f:
            o = json.load(f)
            return GeneratorConfig(**o)


class ModelRewritingConfig(BaseModel):
    do_decompositions_for_model_rewrite: bool
    use_blockwise_compile: bool
    num_blocks_per_supertask: int
    embed_all_constants_into_graph: bool
    add_prefill_last_block_slice: bool = False


class ParallelConfig(BaseModel):
    tensor_parallel_size: int
    pipeline_parallel_size: int

==> furiosa_llm-2024.2.1-py3-none-any.whl: furiosa_llm/models/config_types.py <==
import abc
from enum import Enum
import json
import os
from typing import List, Literal, Optional, Tuple, Union

from pydantic import BaseModel, RootModel
from pydantic.dataclasses import dataclass


class LLMBackend(Enum):
    """The backend implementation to run forward() of a model for the LLM."""

    # FIXME: In order to increase the code consistency, use the capital letter for the enum value
    TORCH_PIPELINE_RUNNER = "torch_pipeline_runner"
    FURIOSA_RT_CUDA = "furiosa_rt_cuda"
    FURIOSA_RT_NPU = "furiosa_rt_npu"
    FURIOSA_RT_V2 = "furiosa_rt_v2"
    MOCK_BACKEND_V2 = "mock_backend_v2"

    def is_parallelism_supported(self) -> bool:
        """Tensor Parallelism/Pipeline Parallelism supports"""
        return self in (
            LLMBackend.TORCH_PIPELINE_RUNNER,
            LLMBackend.FURIOSA_RT_CUDA,
            LLMBackend.FURIOSA_RT_NPU,
            LLMBackend.FURIOSA_RT_V2,
            LLMBackend.MOCK_BACKEND_V2,
        )

    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str):
            return cls[value.upper()]


@dataclass
class SchedulerConfig:
    """
    * version: Version of the `SchedulerConfig`.
    * npu_queue_limit: Maximum number of tasks that can be queued in the hardward
    * max_processing_samples: Maximum number of samples that can be processed by the scheduler
    * spare_blocks_ratio: Ratio of spare blocks that are reserved by scheduler. Smaller value will force the scheduler to use dram aggressively
    * is_offline: If True, use strategies optimzed for offline scenario
    * prefill_chunk_size: Chunk size used for chunked prefill. If the value is `None`, chunked prefill is not used.
    """

    npu_queue_limit: int = 2
    max_processing_samples: int = 65536
    spare_blocks_ratio: float = 0.2
    is_offline: bool = False
    prefill_chunk_size: Optional[int] = None

    # custom comparator to handle float comparison
    def __eq__(self, other):
        return (
            self.npu_queue_limit == other.npu_queue_limit
            and self.max_processing_samples == other.max_processing_samples
            and abs(self.spare_blocks_ratio - other.spare_blocks_ratio) < 1e-6
            and self.is_offline == other.is_offline
        )

    def export(self, path: Union[str, os.PathLike]):
        with open(path, "w") as f:
            f.write(RootModel[SchedulerConfig](self).model_dump_json(indent=4))

    @classmethod
    def load(cls, path: Union[str, os.PathLike]) -> "SchedulerConfig":
        with open(path) as f:
            o = json.load(f)
            return SchedulerConfig(**o)


# FIXME: remove this after fixing generator to support new bucket.
class DeprecatedBucket(BaseModel):
    batch_size: int
    attention_size: int

    def __init__(self, batch_size: int, attention_size: int):
        super(DeprecatedBucket, self).__init__(batch_size=batch_size, attention_size=attention_size)


class Bucket(BaseModel):
    batch_size: int  # batch size, batch size must be a multiple of 2 now.
    attention_size: int
    kv_cache_size: int

    @property
    def input_ids_size(self) -> int:
        return self.attention_size - self.kv_cache_size

    def __init__(self, batch_size: int, attention_size: int, kv_cache_size: int):
        super(Bucket, self).__init__(
            batch_size=batch_size, attention_size=attention_size, kv_cache_size=kv_cache_size
        )

    @classmethod
    def prefill(cls, batch_size: int, attention_size: int) -> "Bucket":
        return cls(batch_size, attention_size, 0)

    @classmethod
    def decode(cls, batch_size: int, attention_size: int) -> "Bucket":
        return cls(batch_size, attention_size, attention_size - 1)

    @property
    def is_prefill(self) -> bool:
        return self.kv_cache_size == 0

    @property
    def is_decode(self) -> bool:
        return self.kv_cache_size > 0

    def __hash__(self) -> int:
        return hash((str(self.__class__), self.batch_size, self.attention_size))


class BucketConfig(abc.ABC): ...


@dataclass
class ManualBucketConfig(BucketConfig):
    prefill_buckets: List[Tuple[int, int]]
    decode_buckets: Optional[List[Tuple[int, int]]] = None


@dataclass
class MinimalBucketConfig(BucketConfig):
    max_seq_len: int


class PagedAttentionConfig(BaseModel):
    """Paged attention configuration.

    Attributes:
        num_blocks (int): The maximum number of blocks that each k/v storage per layer can store
        block_size (int): The maximum number of tokens that can be stored in a single paged attention block
        padding_block_idx (int|None): Padding block's index. This will be used for optimization.
    """

    num_blocks: int
    block_size: int

    # Padding block's index. This will be used for optimization.
    padding_block_idx: Optional[int] = None

    def __init__(self, num_blocks: int, block_size: int, padding_block_idx: Optional[int] = None):
        assert num_blocks > 0
        assert block_size > 0
        super(PagedAttentionConfig, self).__init__(
            num_blocks=num_blocks, block_size=block_size, padding_block_idx=padding_block_idx
        )


class KvCacheSharingAcrossBeamsConfig(BaseModel):
    beam_width: int
    max_new_tokens: int

    def __init__(self, beam_width: int, max_new_tokens: int):
        assert beam_width > 0
        assert max_new_tokens > 0
        super(KvCacheSharingAcrossBeamsConfig, self).__init__(
            beam_width=beam_width, max_new_tokens=max_new_tokens
        )


@dataclass
class GeneratorConfig:
    position_id_pad: int
    buckets: List[Bucket]
    model_qname: str  # qualified name of the model (module + class)
    paged_attention_config: Optional[PagedAttentionConfig]
    packing_type: Literal["IDENTITY"]
    kv_cache_sharing_across_beams_config: Optional[KvCacheSharingAcrossBeamsConfig]
    num_speculative_tokens: Optional[int]

    def export(self, path: Union[str, os.PathLike]):
        with open(path, "w") as f:
            f.write(RootModel[GeneratorConfig](self).model_dump_json(indent=4))

    @classmethod
    def load(cls, path: Union[str, os.PathLike]) -> "GeneratorConfig":
        with open(path) as f:
            o = json.load(f)
            return GeneratorConfig(**o)


class ModelRewritingConfig(BaseModel):
    do_decompositions_for_model_rewrite: bool
    use_blockwise_compile: bool
    num_blocks_per_supertask: int
    embed_all_constants_into_graph: bool
    add_prefill_last_block_slice: bool = False


class ParallelConfig(BaseModel):
    tensor_parallel_size: int
    pipeline_parallel_size: int
